/**************************************************************************************
* Author - Dragomir El Mezeni (based on Altera's example)                             *
* Date - 15/05/2015                                                                   *
*                                                                                     *
* This design example shows how to use the scatter gather DMA (SGDMA) to transfer data*
* from memory to hardware accelerator and other way around.                           *
*                                                                                     *
* This design example performs the following transfers:                               *
*                                                                                     *
* SSRAM(MM) --> (MM)SGDMA(ST) --> (ST)ACCELERATOR(ST) --> (ST)SGDMA(MM) --> SSRAM(MM) *
*                                                                                     *
* Two SGDMA components are used in this system since Avalon-ST is a                   *
* uni-directional point to point method of connecting IP.                             *
***************************************************************************************/
#include "stdio.h"
#include "stdlib.h"
#include "altera_avalon_sgdma.h"
#include "altera_avalon_sgdma_regs.h"
#include "alt_types.h"
#include "sys/alt_timestamp.h"
#include "sys/alt_cache.h"
#include "includes_and_settings.h"  // setup is located in this file

#define MAX_NUMBER_OF_DATA_TRANSFERS 10
#define NUMBER_OF_DATA_TRANSFERS 7

/* These will gate the data checking near the end of main */
volatile alt_u16 tx_done = 0;
volatile alt_u16 rx_done = 0;

void transmit_callback_function(void * context)
{
  tx_done++;  /* main will be polling for this value being 1 */
}

void receive_callback_function(void * context)
{
  rx_done++;  /* main will be polling for this value being 1 */
}

int main()
{
  /* Open a SG-DMA for MM-->ST and ST-->MM (two SG-DMAs are present) */
  alt_sgdma_dev * sgdma_m2s = alt_avalon_sgdma_open("/dev/sgdma_m2s");
  alt_sgdma_dev * sgdma_s2m = alt_avalon_sgdma_open("/dev/sgdma_s2m");

  /* Since descriptors need to be placed in memory locations aligned to
   * descriptor size (which is 4 bytes), larger chunk of memory is first allocated
   * and then aligned location is found. Descriptor pointers point to descriptors
   * placed on aligned location and copy pointer points to entire allocated memory.
   * Copy pointers are needed for properly freeing of allocated memory. */
  alt_sgdma_descriptor *m2s_desc, *m2s_desc_copy;
  alt_sgdma_descriptor *s2m_desc, *s2m_desc_copy;

  alt_u32 return_code;

  alt_16 ** result_buffers;
  alt_8 ** data_buffers;
  alt_u16 * buffer_lengths;

  alt_8 param_a[MAX_NUMBER_OF_DATA_TRANSFERS] = {100, -4, -2, 8, 6,  -7, -15,  100, -5, 10};
  alt_8 param_b[MAX_NUMBER_OF_DATA_TRANSFERS] = {21,  2, -5, 4, 5, -10,  20, -121,  4,  7};

  alt_8 tmp_param_a;
  alt_8 tmp_param_b;

  alt_u32 processing_time_hw[MAX_NUMBER_OF_DATA_TRANSFERS], processing_time_sw[MAX_NUMBER_OF_DATA_TRANSFERS];

  alt_u32 tmp_start_time_hw, tmp_end_time_hw, tmp_start_time_sw, tmp_end_time_sw;

  alt_u32 i, transfer_iteration;

/**************************************************************
   * Making sure the SG-DMAs were opened correctly            *
   ************************************************************/
  if(sgdma_m2s == NULL)
  {
    printf("Could not open the transmit SG-DMA\n");
    return 1; 
  }
  if(sgdma_s2m == NULL)
  {
    printf("Could not open the receive SG-DMA\n");
    return 1;
  }
/**************************************************************/

/**************************************************************
   * Allocating data buffers and populating them with data    *
   ************************************************************/
  return_code = create_test_data(&data_buffers,
                                 &result_buffers,
                                 &buffer_lengths,
                                 NUMBER_OF_BUFFERS,
                                 MINIMUM_BUFFER_LENGTH,
                                 MAXIMUM_BUFFER_LENGTH);
  if(return_code == 1)
  {
    printf("Allocating the data buffers failed... exiting\n");
    return 1;
  }
  

/**************************************************************
   * Allocating descriptor table space from main memory.      * 
   * Pointers are passed by reference since they will be      *
   * modified by this function.                               *
   ************************************************************/
  return_code = create_descriptors(&m2s_desc,
                                &m2s_desc_copy,
                                &s2m_desc,
                                &s2m_desc_copy,
                                data_buffers,
                                result_buffers,
                                buffer_lengths,
                                NUMBER_OF_BUFFERS);
  if(return_code == 1)
  {
    printf("Allocating the descriptor memory failed... exiting\n");
    return 1; 
  }
/**************************************************************/


/**************************************************************
   * Register the ISRs that will get called when each (full)  *
   * transfer completes. When park bit is set, processed      *
   * descriptors are not invalidated (OWNED_BY_HW bit stays 1)*
   * meaning that the same descriptors can be used for new    *
   * transfers.                                               *
   ************************************************************/ 
  alt_avalon_sgdma_register_callback(sgdma_m2s,
                                    &transmit_callback_function,
                                    (ALTERA_AVALON_SGDMA_CONTROL_IE_GLOBAL_MSK | ALTERA_AVALON_SGDMA_CONTROL_IE_CHAIN_COMPLETED_MSK | ALTERA_AVALON_SGDMA_CONTROL_PARK_MSK),
                                    NULL);
  alt_avalon_sgdma_register_callback(sgdma_s2m,
                                    &receive_callback_function,
                                    (ALTERA_AVALON_SGDMA_CONTROL_IE_GLOBAL_MSK | ALTERA_AVALON_SGDMA_CONTROL_IE_CHAIN_COMPLETED_MSK | ALTERA_AVALON_SGDMA_CONTROL_PARK_MSK),
                                    NULL);
/**************************************************************/



/**************************************************************************
   * Data processing with 2 different functions - software and hardware.  *
   * Performance is measured for each approach.                           *
   ************************************************************************/
  for (transfer_iteration = 0; transfer_iteration < NUMBER_OF_DATA_TRANSFERS; transfer_iteration++)
  {
	alt_dcache_flush_all();

	tmp_param_a = param_a[transfer_iteration];
	tmp_param_b = param_b[transfer_iteration];

	alt_timestamp_start();

    tmp_start_time_hw = alt_timestamp();

    return_code = linear_function_hw(sgdma_m2s,
    		                         m2s_desc,
    		                         &tx_done,
    		                         sgdma_s2m,
    		                         s2m_desc,
    		                         &rx_done,
    		                         tmp_param_a,
    		                         tmp_param_b);

    tmp_end_time_hw = alt_timestamp();

    if(return_code == 1)
    {
      printf("Linear function hardware processing failed... exiting\n");
      return 1;
    }

    validate_hw_results(m2s_desc,
    		            s2m_desc,
    		            NUMBER_OF_BUFFERS,
    		            tmp_param_a,
    		            tmp_param_b);

    tmp_start_time_sw = alt_timestamp();

    linear_function_sw(data_buffers,
                       result_buffers,
                       buffer_lengths,
                       NUMBER_OF_BUFFERS,
                       tmp_param_a,
                       tmp_param_b);

    tmp_end_time_sw = alt_timestamp();

    processing_time_hw[transfer_iteration] = tmp_end_time_hw - tmp_start_time_hw;
    processing_time_sw[transfer_iteration] = tmp_end_time_sw - tmp_start_time_sw;

  }
  /**************************************************************/

  for (transfer_iteration = 0; transfer_iteration < NUMBER_OF_DATA_TRANSFERS; transfer_iteration++)
  {
    printf("\nHardware processing duration: %u clock ticks\n", (unsigned int)processing_time_hw[transfer_iteration]);
    printf("Software processing duration: %u clock ticks\n", (unsigned int)processing_time_sw[transfer_iteration]);
  }


/**************************************************************
   * Free allocated memory buffers.						      *
   ************************************************************/
  for (i = 0; i < NUMBER_OF_BUFFERS; i++)
  {
	  free(data_buffers[i]);
	  free(result_buffers[i]);
  }

  free(data_buffers);
  free(result_buffers);

  free(m2s_desc_copy);
  free(s2m_desc_copy);
/**************************************************************/


  printf("Exiting...%c", 4); // 4 will terminate the console

  return 0;
}
